/*
 * kgraft_patch_cve-2015-3339
 *
 * Fix for CVE-2015-3339, bsc#939263
 *
 *  Copyright (c) 2015 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/binfmts.h>
#include <linux/fs.h>
#include <linux/mount.h>

static int (*kgr_security_bprm_set_creds)(struct linux_binprm *bprm);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "security_bprm_set_creds", (void *)&kgr_security_bprm_set_creds },
};

/*
 * New function (CVE-2015-3339)
 */
static void kgr_bprm_fill_uid(struct linux_binprm *bprm)
{
	struct inode *inode;
	unsigned int mode;
	kuid_t uid;
	kgid_t gid;

	/* clear any previous set[ug]id data from a previous binary */
	bprm->cred->euid = current_euid();
	bprm->cred->egid = current_egid();

	if (bprm->file->f_path.mnt->mnt_flags & MNT_NOSUID)
		return;

	if (current->no_new_privs)
		return;

	inode = file_inode(bprm->file);
	mode = ACCESS_ONCE(inode->i_mode);
	if (!(mode & (S_ISUID|S_ISGID)))
		return;

	/* Be careful if suid/sgid is set */
	mutex_lock(&inode->i_mutex);

	/* reload atomically mode/uid/gid now that lock held */
	mode = inode->i_mode;
	uid = inode->i_uid;
	gid = inode->i_gid;
	mutex_unlock(&inode->i_mutex);

	/* We ignore suid/sgid if there are no mappings for them in the ns */
	if (!kuid_has_mapping(bprm->cred->user_ns, uid) ||
		 !kgid_has_mapping(bprm->cred->user_ns, gid))
		return;

	if (mode & S_ISUID) {
		bprm->per_clear |= PER_CLEAR_ON_SETID;
		bprm->cred->euid = uid;
	}

	if ((mode & (S_ISGID | S_IXGRP)) == (S_ISGID | S_IXGRP)) {
		bprm->per_clear |= PER_CLEAR_ON_SETID;
		bprm->cred->egid = gid;
	}
}
/*
 * Patched function (CVE-2015-3339)
 */
int kgr_prepare_binprm(struct linux_binprm *bprm)
{
	int retval;

	if (bprm->file->f_op == NULL)
		return -EACCES;

	kgr_bprm_fill_uid(bprm);

	/* fill in binprm security blob */
	retval = kgr_security_bprm_set_creds(bprm);
	if (retval)
		return retval;
	bprm->cred_prepared = 1;

	memset(bprm->buf, 0, BINPRM_BUF_SIZE);
	return kernel_read(bprm->file, 0, bprm->buf, BINPRM_BUF_SIZE);
}

static int kgr_patch_cve_2015_3339_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int kgr_patch_cve_2015_3339_init(void)
{
	return kgr_patch_cve_2015_3339_kallsyms();
}
